// **************************************************************************
// * This file is part of the FreeFileSync project. It is distributed under *
// * GNU General Public License: http://www.gnu.org/licenses/gpl-3.0        *
// * Copyright (C) Zenju (zenju AT gmx DOT de) - All Rights Reserved        *
// **************************************************************************

#include "version_check.h"
#include <zen/string_tools.h>
#include <zen/i18n.h>
#include <zen/utf.h>
#include <zen/scope_guard.h>
#include <zen/build_info.h>
#include <zen/basic_math.h>
#include <zen/file_error.h>
#include <wx+/popup_dlg.h>
#include "version_id.h"

#ifdef ZEN_WIN
    #include <zen/win.h> //tame wininet.h include
    #include <zen/win_ver.h>
    #include <zen/com_tools.h>
    #include <wininet.h>

#elif defined ZEN_MAC
    #include <CoreServices/CoreServices.h> //Gestalt()
#endif

#if defined ZEN_LINUX || defined ZEN_MAC
    #include <zen/thread.h> //std::thread::id
    #include <wx/protocol/http.h>
    #include <wx/app.h>
#endif


using namespace zen;


namespace
{
#ifndef ZEN_WIN
    #ifndef NDEBUG
        const std::thread::id mainThreadId = std::this_thread::get_id();
    #endif
#endif


std::wstring getIso639Language()
{
    //respect thread-safety for WinInetAccess => don't use wxWidgets in the Windows build here!!!

#ifdef ZEN_WIN //use a more reliable function than wxWidgets:
    const int bufSize = 10;
    wchar_t buf[bufSize] = {};
    int rv = ::GetLocaleInfo(LOCALE_USER_DEFAULT,   //_In_       LCID Locale,
                             LOCALE_SISO639LANGNAME,//_In_       LCTYPE LCType,
                             buf,                   //_Out_opt_  LPTSTR lpLCData,
                             bufSize);              //_In_       int cchData
    if (0 < rv && rv < bufSize)
        return buf; //MSDN: "This can be a 3-letter code for languages that don't have a 2-letter code"!
    assert(false);
    return std::wstring();
#else
    assert(std::this_thread::get_id() == mainThreadId);

    const std::wstring localeName(wxLocale::GetLanguageCanonicalName(wxLocale::GetSystemLanguage()));
    if (localeName.empty())
        return std::wstring();

    assert(beforeLast(localeName, L"_", IF_MISSING_RETURN_ALL).size() == 2);
    return beforeLast(localeName, L"_", IF_MISSING_RETURN_ALL);
#endif
}


std::wstring getIso3166Country()
{
    //respect thread-safety for WinInetAccess => don't use wxWidgets in the Windows build here!!!

#ifdef ZEN_WIN //use a more reliable function than wxWidgets:
    const int bufSize = 10;
    wchar_t buf[bufSize] = {};
    int rv = ::GetLocaleInfo(LOCALE_USER_DEFAULT,    //_In_       LCID Locale,
                             LOCALE_SISO3166CTRYNAME,//_In_       LCTYPE LCType,
                             buf,                    //_Out_opt_  LPTSTR lpLCData,
                             bufSize);               //_In_       int cchData
    if (0 < rv && rv < bufSize)
        return buf; //MSDN: "This can also return a number, such as "029" for Caribbean."!
    assert(false);
    return std::wstring();
#else
    assert(std::this_thread::get_id() == mainThreadId);

    const std::wstring localeName(wxLocale::GetLanguageCanonicalName(wxLocale::GetSystemLanguage()));
    if (localeName.empty())
        return std::wstring();

    return afterLast(localeName, L"_", IF_MISSING_RETURN_NONE);
#endif
}


std::string geHttpPostParameters() //must be in application/x-www-form-urlencoded format!!!
{
    //1. coordinate with get_latest_version_number.php
    //2. respect thread-safety for WinInetAccess => don't use wxWidgets in the Windows build here!!!

    std::string params = "ffs_version=" + utfCvrtTo<std::string>(zen::ffsVersion);

#ifdef ZEN_WIN
    params += "&os_name=Windows";
    const auto osvMajor = getOsVersion().major;
    const auto osvMinor = getOsVersion().minor;

#elif defined ZEN_LINUX
    params += "&os_name=Linux";
    assert(std::this_thread::get_id() == mainThreadId);

    const wxLinuxDistributionInfo distribInfo = wxGetLinuxDistributionInfo();
    assert(contains(distribInfo.Release, L'.'));
    std::vector<wxString> digits = split<wxString>(distribInfo.Release, L'.'); //e.g. "15.04"
    digits.resize(2);
    //distribInfo.Id //e.g. "Ubuntu"

    const int osvMajor = stringTo<int>(digits[0]);
    const int osvMinor = stringTo<int>(digits[1]);

#elif defined ZEN_MAC
    params += "&os_name=Mac";
    SInt32 osvMajor = 0;
    SInt32 osvMinor = 0;
    ::Gestalt(gestaltSystemVersionMajor, &osvMajor);
    ::Gestalt(gestaltSystemVersionMinor, &osvMinor);
#endif
    params += "&os_version=" + numberTo<std::string>(osvMajor) + "." + numberTo<std::string>(osvMinor);

    params +=
#ifdef ZEN_WIN
        running64BitWindows() ? "&os_arch=64" : "&os_arch=32";

#elif defined ZEN_LINUX || defined ZEN_MAC
#ifdef ZEN_BUILD_32BIT
        "&os_arch=32";
#elif defined ZEN_BUILD_64BIT
        "&os_arch=64";
#endif
#endif

    const std::string isoLang    = utfCvrtTo<std::string>(getIso639Language());
    const std::string isoCountry = utfCvrtTo<std::string>(getIso3166Country());
    params += "&language=" + (!isoLang   .empty() ? isoLang    : "zz");
    params += "&country="  + (!isoCountry.empty() ? isoCountry : "ZZ");

    return params;
}


std::string sendHttpRequestImpl(const std::wstring& url, //throw FileError
                                const std::string* postParams, //issue POST if bound, GET otherwise
                                int level = 0)
{
    assert(!startsWith(url, L"https:")); //not supported by wxHTTP!
    std::wstring urlFmt = url;
    if (startsWith(urlFmt, L"http://"))
        urlFmt = afterFirst(urlFmt, L"://", IF_MISSING_RETURN_NONE);
    const std::wstring server =       beforeFirst(urlFmt, L'/', IF_MISSING_RETURN_ALL);
    const std::wstring page   = L'/' + afterFirst(urlFmt, L'/', IF_MISSING_RETURN_NONE);

#ifdef ZEN_WIN
    //WinInet: 1. uses IE proxy settings! :) 2. follows HTTP redirects by default 3. swallows HTTPS if needed
    HINTERNET hInternet = ::InternetOpen(L"FFS-Update-Check",          //_In_  LPCTSTR lpszAgent,
                                         INTERNET_OPEN_TYPE_PRECONFIG, //_In_  DWORD dwAccessType,
                                         nullptr, //_In_  LPCTSTR lpszProxyName,
                                         nullptr, //_In_  LPCTSTR lpszProxyBypass,
                                         0);      //_In_  DWORD dwFlags
    if (!hInternet)
        THROW_LAST_FILE_ERROR(_("Internet access failed."), L"InternetOpen");
    ZEN_ON_SCOPE_EXIT(::InternetCloseHandle(hInternet));

    HINTERNET hSession = ::InternetConnect(hInternet,      //_In_ HINTERNET     hInternet,
                                           server.c_str(), //_In_ LPCTSTR       lpszServerName,
                                           INTERNET_DEFAULT_HTTP_PORT, //_In_ INTERNET_PORT nServerPort,
                                           nullptr,        //_In_ LPCTSTR       lpszUsername,
                                           nullptr,        //_In_ LPCTSTR       lpszPassword,
                                           INTERNET_SERVICE_HTTP, //_In_ DWORD         dwService,
                                           0,              //_In_ DWORD         dwFlags,
                                           0);             //_In_ DWORD_PTR     dwContext
    if (!hSession)
        THROW_LAST_FILE_ERROR(_("Internet access failed."), L"InternetConnect");
    ZEN_ON_SCOPE_EXIT(::InternetCloseHandle(hSession));

    const wchar_t* acceptTypes[] = { L"*/*", nullptr };
    DWORD requestFlags = INTERNET_FLAG_KEEP_CONNECTION |
                         INTERNET_FLAG_NO_UI  |
                         INTERNET_FLAG_RELOAD |                  //
                         INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS; //relevant for GET only
    if (postParams)
        requestFlags |= INTERNET_FLAG_NO_AUTO_REDIRECT; //POST would be re-issued as GET during auto-redirect => handle ourselves!

    HINTERNET hRequest = ::HttpOpenRequest(hSession,     //_In_ HINTERNET hConnect,
                                           postParams ? L"POST" : L"GET", //_In_ LPCTSTR   lpszVerb,
                                           page.c_str(), //_In_ LPCTSTR   lpszObjectName,
                                           nullptr,      //_In_ LPCTSTR   lpszVersion,
                                           nullptr,      //_In_ LPCTSTR   lpszReferer,
                                           acceptTypes,  //_In_ LPCTSTR   *lplpszAcceptTypes,
                                           requestFlags, //_In_ DWORD     dwFlags,
                                           0); //_In_ DWORD_PTR dwContext
    if (!hRequest)
        THROW_LAST_FILE_ERROR(_("Internet access failed."), L"HttpOpenRequest");
    ZEN_ON_SCOPE_EXIT(::InternetCloseHandle(hRequest));

    const std::wstring headers = postParams ? L"Content-type: application/x-www-form-urlencoded" : L"";
    std::string postParamsTmp = postParams ? *postParams : "";
    char* postParamBuf = postParamsTmp.empty() ? nullptr : &*postParamsTmp.begin();
    if (!::HttpSendRequest(hRequest,              //_In_ HINTERNET hRequest,
                           headers.c_str(),       //_In_ LPCTSTR   lpszHeaders,
                           headers.size(),        //_In_ DWORD     dwHeadersLength,
                           postParamBuf,          //_In_ LPVOID    lpOptional,
                           postParamsTmp.size())) //_In_ DWORD     dwOptionalLength
        THROW_LAST_FILE_ERROR(_("Internet access failed."), L"HttpSendRequest");

    DWORD sc = 0;
    {
        DWORD bufLen = sizeof(sc);
        if (!::HttpQueryInfo(hRequest,    //_In_     HINTERNET hRequest,
                             HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, //_In_     DWORD dwInfoLevel,
                             &sc, //_Inout_  LPVOID lpvBuffer,
                             &bufLen,     //_Inout_  LPDWORD lpdwBufferLength,
                             nullptr))    //_Inout_  LPDWORD lpdwIndex
            THROW_LAST_FILE_ERROR(_("Internet access failed."), L"HttpQueryInfo: HTTP_QUERY_STATUS_CODE");
    }

    //http://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection
    if (sc / 100 == 3) //e.g. 301, 302, 303, 307... we're not too greedy since we check location, too!
    {
        if (level < 5) //"A user agent should not automatically redirect a request more than five times, since such redirections usually indicate an infinite loop."
        {
            DWORD bufLen = 10000;
            std::wstring location(bufLen, L'\0');
            if (!::HttpQueryInfo(hRequest, HTTP_QUERY_LOCATION, &*location.begin(), &bufLen, nullptr))
                THROW_LAST_FILE_ERROR(_("Internet access failed."), L"HttpQueryInfo: HTTP_QUERY_LOCATION");
            if (bufLen >= location.size()) //HttpQueryInfo expected to write terminating zero
                throw FileError(_("Internet access failed."), L"HttpQueryInfo: HTTP_QUERY_LOCATION, buffer overflow");
            location.resize(bufLen);

            if (!location.empty())
                return sendHttpRequestImpl(location, postParams, level + 1);
        }
        throw FileError(_("Internet access failed."), L"Unresolvable redirect.");
    }

    if (sc != HTTP_STATUS_OK) //200
        throw FileError(_("Internet access failed."), replaceCpy<std::wstring>(L"HTTP status code %x.", L"%x", numberTo<std::wstring>(sc)));
    //e.g. 404 - HTTP_STATUS_NOT_FOUND

    std::string buffer;
    const DWORD blockSize = 64 * 1024;
    //internet says "HttpQueryInfo() + HTTP_QUERY_CONTENT_LENGTH" not supported by all http servers...
    for (;;)
    {
        buffer.resize(buffer.size() + blockSize);

        DWORD bytesRead = 0;
        if (!::InternetReadFile(hRequest,    //_In_   HINTERNET hFile,
                                &*(buffer.begin() + buffer.size() - blockSize),  //_Out_  LPVOID lpBuffer,
                                blockSize,   //_In_   DWORD dwNumberOfBytesToRead,
                                &bytesRead)) //_Out_  LPDWORD lpdwNumberOfBytesRead
            THROW_LAST_FILE_ERROR(_("Internet access failed."), L"InternetReadFile");

        if (bytesRead < blockSize)
            buffer.resize(buffer.size() - (blockSize - bytesRead)); //caveat: unsigned arithmetics

        if (bytesRead == 0)
            return buffer;
    }

#else
    assert(std::this_thread::get_id() == mainThreadId);
    assert(wxApp::IsMainLoopRunning());

    wxHTTP webAccess;
    webAccess.SetHeader(L"User-Agent", L"FFS-Update-Check");
    webAccess.SetTimeout(10 /*[s]*/); //default: 10 minutes: WTF are these wxWidgets people thinking???

    if (!webAccess.Connect(server)) //will *not* fail for non-reachable url here!
        throw FileError(_("Internet access failed."), L"wxHTTP::Connect");

    if (postParams)
        if (!webAccess.SetPostText(L"application/x-www-form-urlencoded", utfCvrtTo<wxString>(*postParams)))
            throw FileError(_("Internet access failed."), L"wxHTTP::SetPostText");

    std::unique_ptr<wxInputStream> httpStream(webAccess.GetInputStream(page)); //must be deleted BEFORE webAccess is closed
    const int sc = webAccess.GetResponse();

    //http://en.wikipedia.org/wiki/List_of_HTTP_status_codes#3xx_Redirection
    if (sc / 100 == 3) //e.g. 301, 302, 303, 307... we're not too greedy since we check location, too!
    {
        if (level < 5) //"A user agent should not automatically redirect a request more than five times, since such redirections usually indicate an infinite loop."
        {
            const std::wstring newUrl(webAccess.GetHeader(L"Location"));
            if (!newUrl.empty())
                return sendHttpRequestImpl(newUrl, postParams, level + 1);
        }
        throw FileError(_("Internet access failed."), L"Unresolvable redirect.");
    }

    if (sc != 200) //HTTP_STATUS_OK
        throw FileError(_("Internet access failed."), replaceCpy<std::wstring>(L"HTTP status code %x.", L"%x", numberTo<std::wstring>(sc)));

    if (!httpStream || webAccess.GetError() != wxPROTO_NOERR)
        throw FileError(_("Internet access failed."), L"wxHTTP::GetError");

    std::string buffer;
    int newValue = 0;
    while ((newValue = httpStream->GetC()) != wxEOF)
        buffer.push_back(static_cast<char>(newValue));
    return buffer;
#endif
}


bool internetIsAlive() //noexcept
{
#ifdef ZEN_WIN
    //::InternetAttemptConnect(0) -> not working as expected: succeeds even when there is no internet connection!

    HINTERNET hInternet = ::InternetOpen(L"FreeFileSync", //_In_  LPCTSTR lpszAgent,
                                         INTERNET_OPEN_TYPE_PRECONFIG, //_In_  DWORD dwAccessType,
                                         nullptr, //_In_  LPCTSTR lpszProxyName,
                                         nullptr, //_In_  LPCTSTR lpszProxyBypass,
                                         0);      //_In_  DWORD dwFlags
    if (!hInternet)
        return false;
    ZEN_ON_SCOPE_EXIT(::InternetCloseHandle(hInternet));

    //InternetOpenUrl is shortcut for HTTP:GET with InternetConnect + HttpOpenRequest + HttpSendRequest:
    HINTERNET hRequest = ::InternetOpenUrl(hInternet,   //_In_  HINTERNET hInternet,
                                           L"http://www.google.com/", //_In_  LPCTSTR lpszUrl,
                                           nullptr,     //_In_  LPCTSTR lpszHeaders,
                                           0,           //_In_  DWORD dwHeadersLength,
                                           INTERNET_FLAG_KEEP_CONNECTION |
                                           INTERNET_FLAG_NO_UI  |
                                           INTERNET_FLAG_RELOAD |
                                           INTERNET_FLAG_NO_AUTO_REDIRECT, //_In_  DWORD dwFlags,
                                           0);          //_In_  DWORD_PTR dwContext
    //fails with ERROR_INTERNET_NAME_NOT_RESOLVED if server not found => the server-relative part is checked by HTTP_QUERY_STATUS_CODE!!!
    if (!hRequest)
        return false;
    ZEN_ON_SCOPE_EXIT(::InternetCloseHandle(hRequest));

    DWORD sc = 0;
    {
        DWORD bufLen = sizeof(sc);
        if (!::HttpQueryInfo(hRequest,    //_In_     HINTERNET hRequest,
                             HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, //_In_     DWORD dwInfoLevel,
                             &sc, //_Inout_  LPVOID lpvBuffer,
                             &bufLen,     //_Inout_  LPDWORD lpdwBufferLength,
                             nullptr))    //_Inout_  LPDWORD lpdwIndex
            return false;
    }

#else
    const wxString server = L"www.google.com";
    const wxString page   = L"/";

    wxHTTP webAccess;
    webAccess.SetTimeout(10 /*[s]*/); //default: 10 minutes: WTF are these wxWidgets people thinking???

    if (!webAccess.Connect(server)) //will *not* fail for non-reachable url here!
        return false;

    std::unique_ptr<wxInputStream> httpStream(webAccess.GetInputStream(page)); //call before checking wxHTTP::GetResponse()
    const int sc = webAccess.GetResponse();
#endif
    //attention: http://www.google.com/ might redirect to "https" => don't follow, just return "true"!!!
    return sc / 100 == 2 || //e.g. 200
           sc / 100 == 3;   //e.g. 301, 302, 303, 307... when in doubt, consider internet alive!
}

#if 0 //not needed yet: -Wunused-function on OS X
inline
std::string sendHttpGet(const std::wstring& url) //throw FileError
{
    return sendHttpRequestImpl(url, nullptr); //throw FileError
}
#endif

inline
std::string sendHttpPost(const std::wstring& url, const std::string& postParams) //throw FileError
{
    return sendHttpRequestImpl(url, &postParams); //throw FileError
}


enum GetVerResult
{
    GET_VER_SUCCESS,
    GET_VER_NO_CONNECTION, //no internet connection?
    GET_VER_PAGE_NOT_FOUND //version file seems to have moved! => trigger an update!
};

//access is thread-safe on Windows (WinInet), but not on Linux/OS X (wxWidgets)
GetVerResult getOnlineVersion(std::wstring& version)
{
    try
    {
        //harmonize with wxHTTP: get_latest_version_number.php must be accessible without https!!!
        const std::string buffer = sendHttpPost(L"http://www.freefilesync.org/get_latest_version_number.php", geHttpPostParameters()); //throw FileError
        version = utfCvrtTo<std::wstring>(buffer);
        trim(version); //Windows: remove trailing blank and newline
        return version.empty() ? GET_VER_PAGE_NOT_FOUND : GET_VER_SUCCESS; //empty version possible??
    }
    catch (const FileError&)
    {
        return internetIsAlive() ? GET_VER_PAGE_NOT_FOUND : GET_VER_NO_CONNECTION;
    }
}


std::vector<size_t> parseVersion(const std::wstring& version)
{
    std::vector<std::wstring> digits = split(version, FFS_VERSION_SEPARATOR);

    std::vector<size_t> output;
    std::transform(digits.begin(), digits.end(), std::back_inserter(output), [&](const std::wstring& d) { return stringTo<size_t>(d); });
    return output;
}
}


bool zen::haveNewerVersionOnline(const std::wstring& onlineVersion)
{
    std::vector<size_t> current = parseVersion(zen::ffsVersion);
    std::vector<size_t> online  = parseVersion(onlineVersion);

    if (online.empty() || online[0] == 0) //online version string may be "This website has been moved..." In this case better check for an update
        return true;

    return std::lexicographical_compare(current.begin(), current.end(),
                                        online .begin(), online .end());
}


bool zen::updateCheckActive(time_t lastUpdateCheck)
{
    return lastUpdateCheck != getInactiveCheckId();
}


void zen::disableUpdateCheck(time_t& lastUpdateCheck)
{
    lastUpdateCheck = getInactiveCheckId();
}


void zen::checkForUpdateNow(wxWindow* parent, std::wstring& lastOnlineVersion)
{
    std::wstring onlineVersion;
    switch (getOnlineVersion(onlineVersion))
    {
        case GET_VER_SUCCESS:
            lastOnlineVersion = onlineVersion;
            if (haveNewerVersionOnline(onlineVersion))
            {
                switch (showConfirmationDialog(parent, DialogInfoType::INFO, PopupDialogCfg().
                                               setTitle(_("Check for Program Updates")).
                                               setMainInstructions(_("A new version of FreeFileSync is available:")  + L" " + onlineVersion + L"\n\n" + _("Download now?")),
                                               _("&Download")))
                {
                    case ConfirmationButton::DO_IT:
                        wxLaunchDefaultBrowser(L"http://www.freefilesync.org/get_latest.php");
                        break;
                    case ConfirmationButton::CANCEL:
                        break;
                }
            }
            else
                showNotificationDialog(parent, DialogInfoType::INFO, PopupDialogCfg().
                                       setTitle(_("Check for Program Updates")).
                                       setMainInstructions(_("FreeFileSync is up to date.")));
            break;

        case GET_VER_NO_CONNECTION:
            showNotificationDialog(parent, DialogInfoType::ERROR2, PopupDialogCfg().
                                   setTitle(("Check for Program Updates")).
                                   setMainInstructions(_("Unable to connect to www.freefilesync.org.")));
            break;

        case GET_VER_PAGE_NOT_FOUND:
            lastOnlineVersion = L"unknown";
            switch (showConfirmationDialog(parent, DialogInfoType::ERROR2, PopupDialogCfg().
                                           setTitle(_("Check for Program Updates")).
                                           setMainInstructions(_("Cannot find current FreeFileSync version number online. Do you want to check manually?")),
                                           _("&Check")))
            {
                case ConfirmationButton::DO_IT:
                    wxLaunchDefaultBrowser(L"http://www.freefilesync.org/get_latest.php");
                    break;
                case ConfirmationButton::CANCEL:
                    break;
            }
            break;
    }
}


bool zen::shouldRunPeriodicUpdateCheck(time_t lastUpdateCheck)
{
    if (updateCheckActive(lastUpdateCheck))
    {
        const time_t now = std::time(nullptr);
        return numeric::dist(now, lastUpdateCheck) >= 7 * 24 * 3600; //check weekly
    }
    return false;
}


struct zen::UpdateCheckResult
{
#ifdef ZEN_WIN
    GetVerResult versionStatus = GET_VER_PAGE_NOT_FOUND;
    std::wstring onlineVersion;
#endif
};


std::shared_ptr<UpdateCheckResult> zen::retrieveOnlineVersion()
{
#ifdef ZEN_WIN
    try
    {
        ComInitializer ci; //throw SysError

        auto result = std::make_shared<UpdateCheckResult>();
        result->versionStatus = getOnlineVersion(result->onlineVersion); //access is thread-safe on Windows only!
        return result;
    }
    catch (SysError&) { assert(false); return nullptr; }
#else
    return nullptr;
#endif
}


void zen::evalPeriodicUpdateCheck(wxWindow* parent, time_t& lastUpdateCheck, std::wstring& lastOnlineVersion, const UpdateCheckResult* result)
{
#ifdef ZEN_WIN
    const GetVerResult versionStatus = result->versionStatus;
    const std::wstring onlineVersion = result->onlineVersion;
#else
    std::wstring onlineVersion;
    const GetVerResult versionStatus = getOnlineVersion(onlineVersion);
#endif

    switch (versionStatus)
    {
        case GET_VER_SUCCESS:
            lastUpdateCheck = std::time(nullptr);
            lastOnlineVersion = onlineVersion;

            if (haveNewerVersionOnline(onlineVersion))
            {
                switch (showConfirmationDialog(parent, DialogInfoType::INFO, PopupDialogCfg().
                                               setTitle(_("Check for Program Updates")).
                                               setMainInstructions(_("A new version of FreeFileSync is available:")  + L" " + onlineVersion + L"\n\n" + _("Download now?")),
                                               _("&Download")))
                {
                    case ConfirmationButton::DO_IT:
                        wxLaunchDefaultBrowser(L"http://www.freefilesync.org/get_latest.php");
                        break;
                    case ConfirmationButton::CANCEL:
                        break;
                }
            }
            break;

        case GET_VER_NO_CONNECTION:
            break; //ignore this error

        case GET_VER_PAGE_NOT_FOUND:
            lastOnlineVersion = L"unknown";
            switch (showConfirmationDialog(parent, DialogInfoType::ERROR2, PopupDialogCfg().
                                           setTitle(_("Check for Program Updates")).
                                           setMainInstructions(_("Cannot find current FreeFileSync version number online. Do you want to check manually?")),
                                           _("&Check")))
            {
                case ConfirmationButton::DO_IT:
                    wxLaunchDefaultBrowser(L"http://www.freefilesync.org/get_latest.php");
                    break;
                case ConfirmationButton::CANCEL:
                    break;
            }
            break;
    }
}
